/*
	File:		CADoc.cpp

	Contains:	Container Application Library source

	Written by:	Rick Badertscher, Tantek elik, Greg Ames, David Nelson

	Copyright:	 1994 by Apple Computer, Inc., all rights reserved.

	Change History (most recent first):

		<18>	 11/5/95	RB		Serious reconstruction, adding support for new requirements.
									Added multiple document kind support.
		<17>	 9/13/95	TWB		Comment out unused variables. 
		<16>	 6/15/95	RB		Restructure.  Use CADocument.
		<15>	 5/15/95	RB		Adding CA_CATCH
		<14>	  5/8/95	RB		Dumb bug fix
		<13>	  5/6/95	SJF		WWDC bug fix session with Rick
		 <12>	 4/19/95	RB		Mods for b1c18 compliance.
		 <11>	 4/13/95	RB		Added routine GetDraftFromDocumentRef()
		 <10>	 4/08/95	RB		Modified clone method in CASaveDocument()
		 							Added ODFinalReleaseObject() where appropriate.
		 <9>	 4/04/95	RB		Fixes for b1c14.
		 <8>	 4/02/95	RB		Integrating changes for b1c14 build
		 <7>	 3/31/95	RB		Code cleanup.  Use ODDeleteObject(), ODReleaseObject(), etc.
		 <6>	 3/16/95	RB		Changed API to always maintain a docStream SU
		 <5>	 3/15/95	RB		Added support for CA Bento file model.
		 <4+>	 2/28/95	SJF		Remove include of private Container Manager interfaces
		 							"CMDraft.xih & CMCtr.xih, Add ODCtr.xh, mergin in changes
									from RickB, Add debug stuff to API calls
		 <4>	 2/13/95	SJF		Interim checkin to update project database
		 <3)	12/15/94	SJF		change somGetGlobalEnvironment to _gpProxyShell->GetGlobalEnvironment
		 <2>	10/30/94	GCA,DHN	All mods to make project compile with no
		 							errors under OpenDoc b1 (with SOM).
		 <1>	  8/9/94	T		first written
		 
	To Do:
	
	Description:
	
		A CA may specify to save a CADocument to either a file or to a buffer.
		A CA may save a CADocument as a flattened stream or as a OpenDoc container document.
*/

#ifndef _CASESSN_
#include "CASessn.h"
#endif

#ifndef _CAERROR_
#include "CAError.h"
#endif

#ifndef _CADOCPRIV_
#include "CADocPriv.h"
#endif

#ifndef _CAFILEUTILS_
#include "CAFileUtils.h"
#endif

#ifndef _CAUTIL_
#include "CAUtil.h"
#endif

#ifndef SOM_ODDocument_xh
#include <Document.xh>
#endif

#ifndef SOM_ODDraft_xh
#include <Draft.xh>
#endif

#ifndef SOM_ODStorageUnit_xh
#include <StorageU.xh>
#endif

#ifndef SOM_Module_OpenDoc_StdProps_defined
#include <StdProps.xh>
#endif

#ifndef SOM_Module_OpenDoc_StdDefs_defined
#include <StdDefs.xh>
#endif

#ifndef SOM_Module_OpenDoc_StdTypes_defined
#include <StdTypes.xh>
#endif

#ifndef SOM_ODSession_xh
#include <ODSessn.xh>
#endif

#ifndef _PLFMFILE_
#include <PlfmFile.h>
#endif

#ifndef SOM_ODContainer_xh
#include <ODCtr.xh>
#endif

#ifndef _ODUTILS_
#include <ODUtils.h>
#endif

#ifndef _STORUTIL_
#include <StorUtil.h>
#endif

#ifndef _StdTypIO_
#include <StdTypIO.h>
#endif

#ifndef _CAPROXYPARTDEF_
#include "CAProxyPartDef.h"
#endif

#ifndef _INFOUTIL_
#include <InfoUtil.h>
#endif

#ifndef SOM_ODPart_xh
#include <Part.xh>
#endif

#ifndef _DOCUTILS_
#include <DocUtils.h>
#endif

#ifndef _CADEBUG_
#include "CADebug.h"
#endif

#ifndef _ISOSTR_
#include <ISOStr.h>
#endif

#ifndef _PASCLSTR_
#include "PasclStr.h"
#endif


#pragma segment CALib

//-------------------------------------------------------------------------
// Static Utility Routines
//-------------------------------------------------------------------------

// Methods to extract the container stream from CA files or buffers

static PlatformFile*	CreateTempFileFromMemory(	Handle			docHandle,
													CAOffset		offset,
													CASize			length);

static Handle			CreateTempMemoryFromFile(	PlatformFile* 	file,
													CAOffset		offset,
													CASize			length);

static PlatformFile*	CreateTempFileFromFile(		PlatformFile* 	file,
													CAOffset		offset,
													CASize			length);

static Handle			CreateTempMemoryFromMemory(	Handle 			docHandle,
													CAOffset		offset,
													CASize			length);


//===========================================================================
#pragma mark '		Document Management Routines
//===========================================================================

//-------------------------------------------------------------------------
// CACreateDocument()
//-------------------------------------------------------------------------

pascal	CADocumentRef	CACreateDocument(	FSSpecPtr 			docSpec,
											Handle				docHandle)
{
	Environment*			ev				= gCASession->GetEV();
	ODDocument*				newDocument		= kODNULL;
	ODContainer*			newContainer	= kODNULL;
	ODDraft*				newDraft		= kODNULL;
	ODDraft*				baseDraft		= kODNULL;
	ODStorageUnit*			rootPartSU		= kODNULL;
	CADocument*				docRef			= kODNULL;
	PlatformFile*			newFile			= kODNULL;
	ODHandle				ctrHandle		= kODNULL;
	ODPart*					rootPart		= kODNULL;
	Boolean					inPlace			= false;
	
	if (docHandle)
	{
		if (docSpec != kODNULL) SetCAError (kCAErrInvalidParameter);
	}

	CA_TRY
	
		if (docSpec)
		{
		
			// file spec specified, create a bento container in the specified file
			newFile = new PlatformFile;
			newFile->Specify (docSpec);

			if (!newFile->Exists())
			{
				newFile->Create(kODShellSignature, kODShellSignature, 0);
			}

			// Create the container, document, base draft, proxy part
			newContainer = CreateFileContainer(ev, gCASession->GetODSession(), &(newFile->GetFileSpec()));
		
			inPlace = true;
			
		}
		else if (docHandle)
		{
			// handle specified, create a bento container in the specified handle
			ctrHandle = docHandle;
			
			newContainer = CreateMemoryContainer(ev, gCASession->GetODSession(),
								ctrHandle, kODBentoMemoryContainer);
		}
		
		else
		{
			// No file spec or handle specified, create a temporary file to hold
			// the working container
		
			newFile = CreateTempFile(kODShellSignature, kODShellSignature, NULL);
			
			// Create the container
			newContainer = CreateFileContainer(ev, gCASession->GetODSession(), &(newFile->GetFileSpec()));

		}
		
		newDocument = newContainer->AcquireDocument(ev, kODDefaultDocument );
		baseDraft = newDocument->AcquireBaseDraft(ev,kODDPExclusiveWrite);
		rootPart = baseDraft->CreatePart(ev, kCAProxyPartKind, kCAProxyPartID );
		
		ODSetRootPartSUOfDraft(ev, baseDraft, rootPart->GetStorageUnit(ev));
		
		rootPart->Release(ev);
		baseDraft->Externalize(ev);
		
		// Create the working draft
		
		newDraft = newDocument->CreateDraft (ev, baseDraft, kODTrue);
		
		docRef = new CADocument;
		
		if (!docHandle)
		{
			((CADocument*)docRef)->InitCADocument (newDraft, (ODFileSpec*) docSpec,
													newFile, !inPlace, kCAProxyDocument,
													kCATempFileContainer);
		}
		else
		{
			((CADocument*)docRef)->InitCADocument (newDraft, (ODFileSpec*) docSpec,
													ctrHandle, kCAProxyDocument,
													kCATempMemoryContainer);
		}
		
	CA_CATCH_ALL
	
		ODReleaseObject (ev, newContainer);
		ODReleaseObject (ev, newDocument);
		ODReleaseObject (ev, newDraft);
		
	CA_ENDTRY
	
	return ((CADocumentRef)docRef);
	
}


//-------------------------------------------------------------------------
// CAOpenDocument
//
//	If offset & length are 0 then docSpec is presumed to be a container file,
//		the CA is using the Bento storage model
//  If offset & length are not 0 then the CA is using the Native storage model
//		and a ODContainer is located offset bytes into docSpec
//	A temp file containing the provided ODContainer is created.
//
//-------------------------------------------------------------------------
pascal	CADocumentRef	CAOpenDocument( FSSpecPtr			docSpec,
										Handle				docHandle , 
										CAOffset			offset,
										CASize				length)
{

	TRACE (0, "CAOpenDocument - Entry\n");

	Environment*		ev				= gCASession->GetEV();
	ODWindowState*		windowState		= gCASession->GetODSession()->GetWindowState(ev);
	ODDraft*			baseDraft		= kODNULL;
	ODDraft*			topDraft		= kODNULL;
	ODDraft*			newDraft		= kODNULL;
	ODDocument*			newDocument		= kODNULL;
	ODContainer*		newContainer	= kODNULL;
	ODStorageUnit*		rootPartSU		= kODNULL;
	Boolean				deleteFile		= true;
	PlatformFile*		docFile;
	PlatformFile*		containerFile;
	CADocument*			docRef;
	ODHandle			tempCtrHandle;
	CAContainerType		caCtrType;

	if (docHandle)
	{
		if (docSpec != kODNULL) SetCAError (kCAErrInvalidParameter);
	}

	// For both Bento and Native storage models we will create a 
	// temp file to hold the container.  This way a CA can't screw
	// up by mucking with the file
	
	if (docHandle)
	{
	
		CA_TRY
		
			if (offset != 0)
			{
		
				// Create memory container from memory stream
				tempCtrHandle = CreateTempMemoryFromMemory (docHandle, offset, length);
				newContainer = GetMemoryContainer(ev, gCASession->GetODSession(),
									tempCtrHandle, kODDefaultMemoryContainer);
				caCtrType = kCATempMemoryContainer;
						
			}
			else
			{
				// Open memory container in place
				newContainer = GetMemoryContainer(ev, gCASession->GetODSession(),
									docHandle, kODDefaultMemoryContainer);
				tempCtrHandle = docHandle;
				caCtrType = kCAInPlaceMemoryContainer;

			}
			

		CA_CATCH_ALL
		CA_ENDTRY

	}
	
	else
	{
	
		ODContainerType		containerType;

		// Create a container file from the docFile at offset
		
		docFile = new PlatformFile;
		docFile->Specify (docSpec);
		docFile->SetPermission(fsRdPerm);
		
		// If docSpec is part stationary or a part document then no
		// CALib document header bytes will exists. (This is the only case);

		if (docFile->GetPlatformCreator() == kODShellSignature)
		{
					
			ODAcquireCtrDocTopDraft(ev, gCASession->GetODSession(), docFile, &newContainer,
								&newDocument, &newDraft);
	
			docRef = new CADocument ();
			docRef->InitCADocument (newDraft, (ODFileSpec*) docSpec,
											containerFile, false, kCAPartDocument,
											kCAInPlaceFileContainer);
			
			ODDeleteObject (docFile);
			return ((CADocumentRef)docRef);
			
		}
		
		// See if we're dealing with a container file or a flattened container stream
		containerType = docFile->GetContainerType();
		
		if (containerType)
		{
			if ( (offset != 0) || (length != 0) )
				THROW (kCAErrInvalidParameter);
				
			// docFile is a OpenDoc container file, open the container in place
			newContainer = GetFileContainer(ev, gCASession->GetODSession(), docSpec);
			caCtrType = kCAInPlaceFileContainer;
			deleteFile = false;
		}
		else
		{
		
			if ( (offset == 0) || (length == 0) )
				THROW (kCAErrInvalidParameter);

			containerFile = CreateTempFileFromFile (docFile, offset, length);
			newContainer = GetFileContainer(ev, gCASession->GetODSession(),
								&(containerFile->GetFileSpec()));
			caCtrType = kCATempFileContainer;

		}
		
		ODDeleteObject (docFile);

	}
	
	// Now we have a ODContainer for the temp file or memory container
	// Initialize the working draft
	
	newDocument = newContainer->AcquireDocument(ev,kODDefaultDocument);
	baseDraft = newDocument->AcquireBaseDraft (ev, kODDPReadOnly);
	
	CA_TRY
	
		if (IsCAProxyPartDraft (baseDraft))
		{
		
			if (newDocument->Exists (ev, 0, baseDraft, kODPosFirstAbove))
			{
				newDraft = newDocument->AcquireDraft (ev, kODDPExclusiveWrite, 0, baseDraft,
														kODPosFirstAbove, kODTrue);
			}
			else
			{
				newDraft = newDocument->CreateDraft (ev, baseDraft, kODTrue);
			}
			
			// Internalize the window state
			//windowState->Internalize (ev, newDraft);

			docRef = new CADocument ();
			
			if (!docHandle)
				docRef->InitCADocument (newDraft, (ODFileSpec*) docSpec,
												containerFile, deleteFile,
												kCAProxyDocument,
												caCtrType);
			else
				docRef->InitCADocument (newDraft, (ODFileSpec*) docSpec,
												tempCtrHandle, kCAProxyDocument,
												caCtrType);

		}
		else
		{
		
			// Opening a non-proxy container.  Probably a memory based data transfer ctr
			baseDraft->Release(ev);
			newDraft = newDocument->AcquireDraft(ev,kODDPReadOnly,0,kODNULL,kODPosTop,kODFalse);

			docRef = new CADocument ();
			
			if (!docHandle)
				docRef->InitCADocument (newDraft, (ODFileSpec*) docSpec,
												containerFile, false, kCAPartDocument,
												caCtrType);
			else
				docRef->InitCADocument (newDraft, (ODFileSpec*) docSpec,
												tempCtrHandle, kCAPartDocument,
												caCtrType);

		}	
	
	CA_CATCH_ALL
		ODReleaseObject (ev, topDraft);
		ODReleaseObject (ev, baseDraft);
		ODReleaseObject (ev, rootPartSU);
		ODReleaseObject (ev, newDocument);
		ODReleaseObject (ev, newContainer);
		ODDeleteObject (docRef); docRef = NULL;
	CA_ENDTRY

	return ((CADocumentRef)docRef);

}

//-------------------------------------------------------------------------

pascal	Boolean			CAHasDocumentChanged( CADocumentRef document)
{

	Environment*	ev		= gCASession->GetEV();
	ODDraft*		draft	= ((CADocument*)document)->GetDraft();
	Boolean			changed;
	
	CA_TRY
		changed = draft->ChangedFromPrev(ev);
	CA_CATCH_ALL
	CA_ENDTRY
	
	return (changed);
	
}
//-------------------------------------------------------------------------
// CASaveDocument
//-------------------------------------------------------------------------
pascal	void 			CASaveDocument( CADocumentRef	document,
										FSSpecPtr		newFile,
										Handle*			docHandle, 
										CAOffset		offset,
										CASize*			length )
{
	Environment*		ev				= gCASession->GetEV();
	ODWindowState*		windowState		= gCASession->GetODSession()->GetWindowState(ev);
	ODDraft*			theDraft		= kODNULL;
	PlatformFile*		docFile			= kODNULL;
	PlatformFile*		tmpFile			= kODNULL;
	CADocument*			docRef			= kODNULL;
	ODHandle			ctrHandle		= kODNULL;
	ODSLong				count 			= 0;
	const ODSShort		kBufferSize		= 1000 ;
	ODSLong				returnLength	= 0;

	// Validate paramaters
	
	if (docHandle)
	{
		if (newFile != NULL) SetCAError (kCAErrInvalidParameter);
	}
	
	docRef = (CADocument*) document;

	// Externalize the window state
	//windowState->Externalize (ev, docRef->GetDraft());
	
	CA_TRY
	
		if (!docHandle && !newFile)
		{
			// Validate parameters
			if (offset != 0)
				THROW (kCAErrInvalidParameter);

			// Save the container in place
			returnLength = (docRef)->Save ();
		}

		if (newFile)
		{
		
			docFile = new PlatformFile;
			docFile->Specify (newFile);

			// If newFile == docRef->GetCAFileSpec then save the file in place
			
			if ( (newFile->parID == ((docRef)->GetCAFileSpec()).parID) &&
				 (newFile->vRefNum == ((docRef)->GetCAFileSpec()).vRefNum) &&
				 (EqualPascalStrings(newFile->name, ((docRef)->GetCAFileSpec()).name)))
			{
				// Validate parameters
				if (offset != 0)
				{
				
					// Stream the container to newFile at offset

					docFile->SetPermission(fsRdWrPerm);
					docFile->Open ();
					docFile->SetFilePos (fsFromStart, offset);
	
					// Dump the CADocument to the file (Creates an intermediate container
					// file and streams it into the CA specified file)
					returnLength = docRef->StreamToFile (docFile);

					docFile->Close();
					
					//THROW (kCAErrInvalidParameter);

				}
				else
				{
					// Save the container in place
					returnLength = (docRef)->Save ();
				}
				
			}
			else
			{
				if ((offset == 0))
				{

					if (docRef->GetContainerType() == kCAInPlaceFileContainer)
						DebugStr ("\pTrying to save a copy of an in place document");
						
					returnLength = docRef->SaveAsFile (docFile);
					
					// Save a copy of the container file in newFile
					
					ODContainerType		type;
	
					type = (ODContainerType) ODNewPtr (ODISOStrLength(kODDefaultFileContainer) + 1);
					ODISOStrCopy (type, kODDefaultFileContainer);
					docFile->SetContainerType (type);
					
				}
				else
				{
					// Stream the container to newFile at offset

					docFile->SetPermission(fsRdWrPerm);
					docFile->Open ();
					docFile->SetFilePos (fsFromStart, offset);
	
					// Dump the CADocument to the file (Creates an intermediate container
					// file and streams it into the CA specified file)
					returnLength = docRef->StreamToFile (docFile);

					docFile->Close();
				}
				
			}
		}
		else if (docHandle)
		{
			// Validate parameters
			//if ((offset == 0) && (length == NULL))
			//{
				// Save the container in place
			//	*length = docRef->Save ();
			//}
			if ((docRef->GetContainerType() == kCATempFileContainer) ||
			    (docRef->GetContainerType() == kCATempFileContainer))
			{
			
				DebugStr ("\pTrying to save a file container to memory");
			}
			else // Memory container
			{
				if (offset == 0)
				{
					// Saves a copy of hte memory container in docHandle
					CloneDraftToMemoryContainer	(docRef->GetDraft(), docHandle);
					returnLength = ODGetHandleSize (*docHandle);

				}
				else
				{
					// Stream the CADocument container to memory at offset
					returnLength = ((CADocument*)docRef)->StreamToMemory (docHandle, offset);
				}
			}
		}

	CA_CATCH_ALL
	CA_ENDTRY

	if (length) *length = returnLength;
	
	ODDeleteObject (docFile);
	

}


//-------------------------------------------------------------------------
// CACloseDocument
//
//	Releases the temp file associated with the documentRef.
//	document is no longer valid
//-------------------------------------------------------------------------

pascal	void 			CACloseDocument( CADocumentRef document)
{
	TRACE (0, "CACloseDocument - Entry");

	Environment*	ev				= gCASession->GetEV();
	ODStorageUnit*	rootPartSU		= kODNULL;
	CADocument*		docRef			= kODNULL;
	
	docRef = (CADocument*) document;

	CA_TRY

		if (docRef->Close())	// returns true if we should delete it
			ODDeleteObject (docRef);

	
	CA_CATCH_ALL
	CA_ENDTRY
	
}

//-------------------------------------------------------------------------
// CASetDocumentKind - Sets value kind to the currently focused value.
//					   
//
pascal	void			CASetDocumentKind( CADocumentRef document, CAISOStr kind )
{
	Environment*			ev = gCASession->GetEV();
	ODStorageUnit*			rootSU;
	ODISOStr				currentKind;
	
	rootSU = ((CADocument*)document)->GetRootPartSU();

	CA_TRY
	
		if (!rootSU) THROW (kCAErrCantFocusDocument);

		// If there already exists a value, then change it's type
		if (rootSU->Exists (ev, kODPropContents, kODNULL, 1))
		{
			// Unless it's the same the existing value is kind, then just 
			// set it as the current kind
			rootSU->Focus (ev, kODPropContents, kODPosSame, kODNULL, 1, kODPosUndefined);
			currentKind = rootSU->GetType(ev);

			if (ODISOStrCompare (kind, currentKind))
				rootSU->SetType (ev, kind);
		}
		else
		{
			ODSUForceFocus(ev, rootSU, kODPropContents, kind);
		}
		
	CA_CATCH_ALL
	CA_ENDTRY

	((CADocument*)document)->SetStreamOffset (0);
	
}


//-------------------------------------------------------------------------

pascal	CAISOStr CAGetDocumentKind( CADocumentRef document)
{

	Environment*		ev				= gCASession->GetEV();
	ODISOStr			valueType		= NULL;
	CAISOStr			appMemoryStr	= NULL;
	ODStorageUnit*		rootSU;

	rootSU = ((CADocument*)document)->GetRootPartSU();

	CA_TRY
	
		if (!rootSU) THROW (kCAErrCantFocusDocument);
		
		if (rootSU->Exists (ev, kODPropContents, kODNULL, 1))
			valueType = rootSU->GetType (ev);
		
		if (valueType)
		{
			appMemoryStr = (CAISOStr) NewPtr (ODISOStrLength (valueType) + 1);
			ODISOStrCopy (appMemoryStr, valueType);
			ODDisposePtr (valueType);
		}
	
	CA_CATCH_ALL
	CA_ENDTRY
	
	return ((CAISOStr) appMemoryStr);
}


//===========================================================================
#pragma mark '		Document Kind Routines
//===========================================================================

//-------------------------------------------------------------------------

pascal void		CAAddKind (CADocumentRef document, CAISOStr kind)
{

	Environment*		ev			= gCASession->GetEV();
	ODStorageUnit*		rootSU		= kODNULL;

	rootSU = ((CADocument*)document)->GetRootPartSU();
	
	CA_TRY
	
		if (!rootSU) THROW (kCAErrCantFocusDocument);
		
		// If it's already present flag an error, but set current kind anyway
		if (rootSU->Exists(ev, kODPropContents, kind, 0))
		{
			SetCAError (kCAErrKindExists);
			ODSUForceFocus (ev, rootSU, kODPropContents, kind);
		}
		else
		{
			rootSU->Focus(ev,  kODPropContents, kODPosSame, kODNULL, 0, kODPosAll);
			ODSUForceFocus (ev, rootSU, kODPropContents, kind);
		}
		
		((CADocument*)document)->SetCurrentKind (kind);
		
	CA_CATCH_ALL
	CA_ENDTRY
	

}

//-------------------------------------------------------------------------

pascal void		CARemoveKind (CADocumentRef document, CAISOStr kind)
{

	Environment*		ev			= gCASession->GetEV();
	ODStorageUnit*		rootSU		= kODNULL;
	Boolean				exists;
	
	rootSU = ((CADocument*)document)->GetRootPartSU();
	
	CA_TRY
	
		if (!rootSU) THROW (kCAErrCantFocusDocument);

		exists = rootSU->Exists(ev, kODPropContents, kind, 0);

		if (!exists) THROW (kCAErrInvalidKind);

		// Focus on the desired index and get it's type
		rootSU->Focus(ev, kODPropContents, kODPosUndefined, 
						kind, 0, kODPosUndefined);
				
		// Remove it		
		rootSU->Remove (ev);
		
		// Default to first value
		((CADocument*)document)->SetCurrentKind(kODNULL);

		// Refocus the storage unit
		((CADocument*)document)->FocusContentStream();
			
		
	CA_CATCH_ALL
	CA_ENDTRY
	

}


//-------------------------------------------------------------------------

pascal CAISOStr	CAGetNthKind(CADocumentRef document, short index)
{

	Environment*		ev			= gCASession->GetEV();
	ODStorageUnit*		rootSU		= kODNULL;
	Boolean				exists		= false;
	ODISOStr			kind		= kODNULL;
	CAISOStr			appMemoryStr;
	
	rootSU = ((CADocument*)document)->GetRootPartSU();
	
	CA_TRY
	
		if (!rootSU) THROW (kCAErrCantFocusDocument);

		exists = rootSU->Exists(ev, kODPropContents, kODNULL, index);
		
		if (!exists) THROW (kCAErrInvalidKind);
		
		// Focus on the desired index and get it's type
		rootSU->Focus(ev, kODPropContents, kODPosUndefined, 
						kODNULL, index, kODPosUndefined);
						
		kind = rootSU->GetType (ev);
		
		// Refocus the storage unit
		((CADocument*)document)->FocusContentStream();
		
	CA_CATCH_ALL
	CA_ENDTRY
	
	appMemoryStr = (CAISOStr) NewPtr (ODISOStrLength (kind) + 1);
	ODISOStrCopy (appMemoryStr, kind);
	ODDisposePtr (kind);
	
	return ((CAISOStr) appMemoryStr);

	return (kind);

}

//-------------------------------------------------------------------------

pascal short	CAGetKindCount (CADocumentRef document)
{

	Environment*		ev			= gCASession->GetEV();
	ODStorageUnit*		rootSU		= kODNULL;
	short				count;

	rootSU = ((CADocument*)document)->GetRootPartSU();
	
	CA_TRY
	
		if (!rootSU) THROW (kCAErrCantFocusDocument);
		count = rootSU->CountValues(ev);
		
	CA_CATCH_ALL
	CA_ENDTRY
	
	return (count);

}


//-------------------------------------------------------------------------

pascal Boolean CAUseKind(CADocumentRef document, CAISOStr kind, short index)
{
	Environment*		ev			= gCASession->GetEV();
	ODStorageUnit*		rootSU		= kODNULL;
	Boolean				exists		= false;
	ODValueType			value;
	
	rootSU = ((CADocument*)document)->GetRootPartSU();

	CA_TRY
	
		if (!rootSU) THROW (kCAErrCantFocusDocument);

		if (index)
		{
			if (kind) THROW (kCAErrInvalidParameter);
		}
	
		// Focus the storage unit
		if (index)
		{
			
			exists = rootSU->Exists(ev, kODPropContents, kODNULL, index);

			if (exists)
				rootSU->Focus (ev, kODPropContents, kODPosSame, kODNULL, index, kODPosUndefined);
			
		}
		else
		{
			exists = rootSU->Exists(ev, kODPropContents, kind, 0);

			if (exists)
				rootSU->Focus (ev, kODPropContents, kODPosSame, kind, 0, kODPosUndefined);

		}
		
		if (exists)
		{
			value = rootSU->GetType (ev);
			((CADocument*)document)->SetCurrentKind (value);
			ODDisposePtr (value);
		}

		
	CA_CATCH_ALL
	CA_ENDTRY

	return (exists);

}

//-------------------------------------------------------------------------

pascal CAISOStr CACurrentKind(CADocumentRef document)
{
	Environment*		ev				= gCASession->GetEV();
	ODStorageUnit*		rootSU			= kODNULL;
	ODISOStr			valueType		= kODNULL;
	CAISOStr			appMemoryStr	= kODNULL;

	rootSU = ((CADocument*)document)->GetRootPartSU();
	
	CA_TRY
	
		if (!rootSU) THROW (kCAErrCantFocusDocument);

		valueType = ((CADocument*)document)->GetCurrentKind();
	
		if (valueType)
		{
			appMemoryStr = (CAISOStr) NewPtr (ODISOStrLength (valueType) + 1);
			ODISOStrCopy (appMemoryStr, valueType);
			ODDisposePtr (valueType);
		}
	
	CA_CATCH_ALL
	CA_ENDTRY
	
	return ((CAISOStr) appMemoryStr);

}

//===========================================================================
#pragma mark '		Utility Routines
//===========================================================================

//-------------------------------------------------------------------------

Boolean IsCAProxyPart (ODPart* part)
{
	Environment*		ev					= gCASession->GetEV();
	Boolean				retVal				= false;
	ODType				kind;
	
	kind = ODGetKindFromPartSU(ev, part->GetStorageUnit(ev));
	
	if (!ODISOStrCompare (kind, kCAProxyPartKind))
	{
		retVal = true;
	}
	
	ODDisposePtr( kind );
	
	return (retVal);

}


//-------------------------------------------------------------------------

static PlatformFile*	CreateTempFileFromMemory (	Handle			docHandle,
													CAOffset		offset,
													CASize			length)

{
	Environment*		ev					= gCASession->GetEV();
	const ODSShort		kBufferSize			= 1000 ;
	ODSLong				count;
	PlatformFile*		tempFile;

	// Create the temp file

	tempFile = CreateTempFile(kODShellSignature,
							kODShellSignature, NULL);
	tempFile->SetPermission (fsRdWrPerm);
	
	tempFile->Open ();
	tempFile->SetFilePos (fsFromStart, 0);
			
	//count = offset ? (length - sizeof (ODULong)) : length;	// minus the header
	count = length;
		
	HLock (docHandle);
	tempFile->Write( *docHandle, &count );
	HUnlock (docHandle);

	tempFile->SetEndOfFile (length);
	tempFile->Close();

	return (tempFile);
	
}


//-------------------------------------------------------------------------
static PlatformFile*	CreateTempFileFromFile (PlatformFile* 	docFile,
												CAOffset		offset,
												CASize			length)
{
	Environment*		ev					= gCASession->GetEV();
	const ODSShort		kBufferSize			= 1000 ;
	ODSLong				bytesLeft;
	ODSByte				buffer[kBufferSize];
	ODSLong				count;
	ODSLong				ctrLength;
	PlatformFile*		tempFile;

	// Create the temp file

	tempFile = CreateTempFile(kODShellSignature,
							kODShellSignature, NULL);
							
	tempFile->Rename ( (docFile->GetFileSpec()).name);
	tempFile->SetPermission (fsRdWrPerm);
	
	if ((offset == 0) && (length == 0))
	{
		ODContainerType		srcCtrType, dstCtrType;
		
		// Copy the file contents
		tempFile->CopyFrom (docFile);
		
		// Copy the container type
		srcCtrType = docFile->GetContainerType();
		dstCtrType = (ODContainerType) ODNewPtr (ODISOStrLength(srcCtrType) + 1);
		ODISOStrCopy (dstCtrType, srcCtrType);
		tempFile->SetContainerType (dstCtrType);
		
	}
	else
	{
	
		docFile->Open ();
		//docFile->SetFilePos (fsFromStart, offset ? offset + sizeof (ODULong) : 0);
		docFile->SetFilePos (fsFromStart, offset);
					
		tempFile->Open ();
		tempFile->SetFilePos (fsFromStart, 0);
		
		//ctrLength = offset ? (length - sizeof (ODULong)) : length;	// minus the header
		ctrLength = length;	// minus the header
		bytesLeft = ctrLength;
				
		while( bytesLeft > 0 )
		{
			count = bytesLeft < kBufferSize ? bytesLeft: kBufferSize ;
			docFile->Read( buffer, &count );
			tempFile->Write( buffer, &count );
			bytesLeft -= count ;
		}
		
		tempFile->SetEndOfFile (ctrLength);
		tempFile->Close();
		
		docFile->Close();

	}
	
	return (tempFile);
	
}


//-------------------------------------------------------------------------
static Handle	CreateTempMemoryFromMemory (Handle 		docHandle,
											CAOffset	offset,
											CASize		length)
{
	Environment*		ev					= gCASession->GetEV();
	ODHandle			tempHandle;
	CASize				count;
	ODPtr				srcPtr;
	
	//count = offset ? (length - sizeof (ODULong)) : length;	// minus the header
	count = length;	// minus the header

	tempHandle = ODNewHandle (count);
	THROW_IF_NULL (tempHandle);
	
	ODLockHandle (tempHandle);
	HLock (docHandle);
	srcPtr = *docHandle + offset;
	
	ODBlockMove (srcPtr, *tempHandle, count);
	
	ODUnlockHandle (tempHandle);
	HUnlock (docHandle);

	return (tempHandle);

}

//-------------------------------------------------------------------------
static Handle	CreateTempMemoryFromFile(	PlatformFile*	docFile,
											CAOffset		offset,
											CASize			length)
{
	Environment*		ev					= gCASession->GetEV();
	const ODSShort		kBufferSize			= 1000 ;
	ODHandle			tempHandle;
	ODSLong				count;
	ODPtr				dstPtr;
	ODSLong				bytesLeft;
	ODSByte				buffer[kBufferSize];

	// Create the temp memory
	
	//count = offset ? (length - sizeof (ODULong)) : length;	// minus the header
	count = length;	// minus the header

	tempHandle = ODNewHandle (count);
	THROW_IF_NULL (tempHandle);

	ODLockHandle (tempHandle);
	dstPtr = *tempHandle;
	
	docFile->Open ();
	//docFile->SetFilePos (fsFromStart, offset ? offset + sizeof (ODULong) : 0);
	docFile->SetFilePos (fsFromStart, offset);
				
	//bytesLeft = offset ? (length - sizeof (ODULong)) : length;	// minus the header
	bytesLeft = length;	// minus the header
			
	while( bytesLeft > 0 )
	{
		count = bytesLeft < kBufferSize ? bytesLeft: kBufferSize ;
		docFile->Read( buffer, &count );
		ODBlockMove (buffer, dstPtr, count);
		dstPtr = (Ptr) ((long) dstPtr + count);
		bytesLeft -= count ;
	}
	
	docFile->Close();

	ODUnlockHandle (tempHandle);
	
	return (tempHandle);

}

