/*
	File:		CADocPriv.cpp

	Contains:	CADocument class implementation

	Written by:	Rick Badertscher
	Copyright:	 1994 by Apple Computer, Inc., all rights reserved.

	Change History (most recent first):

		 <1>	11/6/95 RB		Added multiple document kind support.
		 <0>	8/17/94	RB		Created
	To Do:
*/

#ifndef _CADOCPRIV_
#include "CADocPriv.h"
#endif

#ifndef _CASESSN_
#include "CASessn.h"
#endif

#ifndef _CADEBUG_
#include "CADebug.h"
#endif

#ifndef _CAFRMUTL_
#include "CAFrmUtl.h"
#endif

#ifndef _CAFILEUTILS_
#include "CAFileUtils.h"
#endif

#ifndef SOM_ODStorageUnit_xh
#include <StorageU.xh>
#endif

#ifndef SOM_ODFrame_xh
#include <Frame.xh>
#endif

#ifndef SOM_ODFacet_xh
#include <Facet.xh>
#endif

#ifndef SOM_ODFrameFacetIterator_xh
#include <FrFaItr.xh>
#endif

#ifndef SOM_Module_OpenDoc_StdProps_defined
#include <StdProps.xh>
#endif

#ifndef SOM_Module_OpenDoc_StdTypes_defined
#include <StdTypes.xh>
#endif

#ifndef SOM_Module_OpenDoc_StdDefs_defined
#include <StdDefs.xh>
#endif

#ifndef _ODUTILS_
#include <ODUtils.h>
#endif

#ifndef _ORDCOLL_
#include <OrdColl.h>
#endif

#ifndef SOM_ODPart_xh
#include <Part.xh>
#endif

#ifndef _STORUTIL_
#include <StorUtil.h>
#endif

#ifndef _StdTypIO_
#include <StdTypIO.h>
#endif

#ifndef SOM_ODDocument_xh
#include <Document.xh>
#endif

#ifndef SOM_ODDraft_xh
#include <Draft.xh>
#endif

#ifndef SOM_ODContainer_xh
#include <ODCtr.xh>
#endif

#ifndef SOM_ODWindowState_xh
#include <WinStat.xh>
#endif

#ifndef SOM_ODClipboard_xh
#include "Clipbd.xh"
#endif

#ifndef SOM_ODFoci_xh
#include <Foci.xh>
#endif

#ifndef SOM_ODFacetIterator_xh
#include <FacetItr.xh>
#endif

#ifndef SOM_ODStorageUnitRefIterator_xh
#include <SURefItr.xh>
#endif

#ifndef SOM_ODDragAndDrop_xh
#include "DragDrp.xh"
#endif

#ifndef _DRAGPRIV_
#include "DragPriv.h"
#endif

#ifndef _DOCUTILS_
#include <DocUtils.h>
#endif

#ifndef _STORUTIL_
#include <StorUtil.h>
#endif

#ifndef _ISOSTR_
#include <ISOStr.h>
#endif

#ifndef _TEMPOBJ_
#include <TempObj.h>
#endif


#pragma segment CALib


//-------------------------------------------------------------------------
CADocument::CADocument (void)
{

	fStreamOffset		= 0;
	fRootPartSU			= kODNULL;
	fCurrentKind		= kODNULL;
	fLimboFrames		= kODNULL;
}

//-------------------------------------------------------------------------
CADocument::~CADocument ()
{
				
}

//-------------------------------------------------------------------------
void CADocument::InitCADocument(ODDraft*			draft,
								const ODFileSpec*	fileSpec,
								PlatformFile*		tempFile,
								Boolean				deleteFile,
								CADocumentType		docType,
								CAContainerType		ctrType)
{
	Environment*			ev = gCASession->GetEV();
	
	fStreamOffset			= 0;
	fTmpCtrFile 			= tempFile;
	fTmpCtrHandle			= kODNULL;
	fDeleteFile				= deleteFile;
	fDocType				= docType;
	fCtrType				= ctrType;
	fCurrentKind			= kODNULL;
	
	SetWindow(kODNULL);
	SetCAFileSpec(fileSpec);

	TempODStorageUnit tempPropsSU = draft->AcquireDraftProperties(ev);
	fRootPartSU = draft->AcquireStorageUnit (ev, ODGetStrongSURefProp(ev, tempPropsSU,
				kODPropRootPartSU, kODStrongStorageUnitRef));

	// Initialize the current kind
	
	if (docType == kCAProxyDocument)
	{
		if (fRootPartSU->Exists (ev, kODPropContents, kODNULL, 1))
			fRootPartSU->Focus (ev, kODPropContents, kODPosSame, kODNULL, 1, kODPosUndefined);

	}
	
	gCASession->AddCADocumentRef (this);

}

//-------------------------------------------------------------------------
void CADocument::InitCADocument(ODDraft*			draft,
								const ODFileSpec*	fileSpec,
								ODHandle			tempHandle,
								CADocumentType		docType,
								CAContainerType		ctrType)
{
	Environment*			ev = gCASession->GetEV();

	fStreamOffset			= 0;
	fTmpCtrFile 			= kODNULL;
	fTmpCtrHandle			= tempHandle;
	fDeleteFile				= false;
	fDocType				= docType;
	fCtrType				= ctrType;
	fCurrentKind			= kODNULL;

	SetWindow(kODNULL);
	SetCAFileSpec(fileSpec);

	TempODStorageUnit tempPropsSU = draft->AcquireDraftProperties(ev);
	fRootPartSU = draft->AcquireStorageUnit (ev, ODGetStrongSURefProp(ev, tempPropsSU,
				kODPropRootPartSU, kODStrongStorageUnitRef));

	// Initialize the current kind
	
	if (docType == kCAProxyDocument)
	{
		if (fRootPartSU->Exists (ev, kODPropContents, kODNULL, 1))
			fRootPartSU->Focus (ev, kODPropContents, kODPosSame, kODNULL, 1, kODPosUndefined);

	}

	gCASession->AddCADocumentRef (this);

}

//-------------------------------------------------------------------------

Boolean CADocument::Close(void)
{
	Environment*		ev = gCASession->GetEV();
	ODDocument*			theDocument;
 	ODContainer*		theContainer;
	ODDraft*			theDraft;
	ODClipboard*		clipboard;
	
	theDraft = this->GetDraft();

	// Tell the clipboard this draft is closing
	
	if (GetDocumentType() == kCAProxyDocument)
	{
		clipboard = gCASession->GetODSession()->GetClipboard(ev);
		clipboard->DraftClosing(ev, theDraft);
	}
	
	// Clean limbo frame list
	
	if (fLimboFrames)
	{
		fLimboFrames->RemoveAll();
		ODDeleteObject (fLimboFrames);
	}
	
	// Do the releasing
	
	theDocument = theDraft->GetDocument (ev);
	theContainer = theDocument->GetContainer(ev);
	
	ODReleaseObject (ev, fRootPartSU);
	
	ODRelease (ev, theDraft);
	ODRelease (ev, theDocument);
	ODRelease (ev, theContainer);

	// Delete the temp file or handle if neccessery
	
	if (fDeleteFile && fTmpCtrFile)
	{
		fTmpCtrFile->Delete();
		delete fTmpCtrFile;
	}
	
	if (this->GetContainerType() == kCATempMemoryContainer)
	{
		ODDisposeHandle (fTmpCtrHandle);
	}
	
	gCASession->RemoveCADocumentRef (this);

	fWindow = NULL;

	return (true);
	
}

//-------------------------------------------------------------------------
CACloneKind CADocument::GetCloneKind(void)
{
	return kCACloneAll;
}


//-------------------------------------------------------------------------

void CADocument::FocusContentStream(void)
{
	Environment*		ev = gCASession->GetEV();

	if (fCurrentKind)
	{
		// CA using multiple kinds
		if (fRootPartSU->Exists (ev, kODPropContents, fCurrentKind, 0))
		{
			fRootPartSU->Focus(ev, kODPropContents, kODPosUndefined, 
								fCurrentKind, 0, kODPosUndefined);
	
			fRootPartSU->SetOffset (ev, fStreamOffset);
	
		}
		else
			THROW (kCAErrCantFocusDocument);

	}
	else
	{
	
		// CA using single kind
		if (fRootPartSU->Exists (ev, kODPropContents, kODNULL, 1))
		{
			fRootPartSU->Focus(ev, kODPropContents, kODPosUndefined, 
								kODNULL, 1, kODPosUndefined);
	
			fRootPartSU->SetOffset (ev, fStreamOffset);
			
			long streamLength = fRootPartSU->GetSize(ev);
			ODISOStr valueKind;
			valueKind = fRootPartSU->GetType(ev);
			ODDisposePtr (valueKind);

	
		}
		else
			THROW (kCAErrCantFocusDocument);
	
	}
	
}


//-------------------------------------------------------------------------

void CADocument::SetCurrentKind (ODISOStr kind)
{

	if (fCurrentKind) ODDisposePtr (fCurrentKind);
	
	if (kind)
	{
		fCurrentKind = (ODISOStr) ODNewPtr (ODISOStrLength (kind) + 1);
		ODISOStrCopy (fCurrentKind, kind);
	}
	else
		fCurrentKind = kODNULL;
			
	fStreamOffset = 0;
}


//-------------------------------------------------------------------------

ODStorageUnit*	CADocument::GetRootPartSU(void)
{
	return (fRootPartSU);
}

//-------------------------------------------------------------------------

ODDraft*	CADocument::GetDraft(void)
{
	Environment*		ev = gCASession->GetEV();
	return fRootPartSU->GetDraft(ev);
}


//-------------------------------------------------------------------------

ODPart* CADocument::AcquireRootPart(void)
{
	Environment*		ev = gCASession->GetEV();
	ODPart*				rootPart;
	
	rootPart = this->GetDraft()->AcquirePart (ev, fRootPartSU->GetID(ev));
	
	return (rootPart);

}

//-------------------------------------------------------------------------

ODFrame* CADocument::AcquireDisplayFrame(void)
{
	Environment*		ev = gCASession->GetEV();
	ODFrame*			frame = kODNULL;
	ODID				frameID;
	
	if ( fRootPartSU->Exists(ev, kODPropDisplayFrames,
			kODWeakStorageUnitRefs, 0) )
	{
	
		frameID = ODGetWeakSURefProp(ev, fRootPartSU,
					kODPropDisplayFrames, kODWeakStorageUnitRefs);
		
		if (frameID)
		{
			frame = this->GetDraft()->AcquireFrame (ev, frameID);
		}
								
	}
	
	return (frame);

}

//-------------------------------------------------------------------------
CASize CADocument::Save(void)
{
	Environment*		ev = gCASession->GetEV();
	CASize				ctrLength;

	if (this->GetDraft()->ChangedFromPrev(ev))
	{
		// Externalize & Empty the working draft
		this->GetDraft()->Externalize(ev);
		this->GetDraft()->SaveToAPrevious (ev, kODNULL);
	}
	
	if (this->GetFile())
	{
		PlatformFile*	file;
		
		file = new PlatformFile;
		file->SpecifyFromFile (this->GetFile());
		file->SetPermission (fsRdPerm);

		// Get the file length
		file->Open();
		ctrLength = file->GetEndOfFile ();
		file->Close();
		
		ODDeleteObject (file);
	}
	else
	{
		ctrLength = GetHandleSize (this->fTmpCtrHandle);
	}
		
	return (ctrLength);

}


//-------------------------------------------------------------------------
CASize CADocument::SaveAsFile(PlatformFile*	file)
{
	Environment*		ev = gCASession->GetEV();
	CASize				fileLength;
	OSType				creator;
	
	if (this->GetDraft()->ChangedFromPrev(ev))
	{
		// Externalize & Empty the working draft
		this->GetDraft()->Externalize(ev);
		this->GetDraft()->SaveToAPrevious (ev, kODNULL);
	}
	
	creator = file->GetPlatformCreator();
	CloneDraftToFileContainer (this->GetDraft(), file);

	// Get the file length
	file->Open();
	fileLength = file->GetEndOfFile ();
	file->Close();

	return (fileLength);

}


//-------------------------------------------------------------------------

CASize CADocument::StreamToFile(PlatformFile* file)
{
	Environment*		ev = gCASession->GetEV();
	ODSLong				length;
	PlatformFile*		tempFile;
	ODSLong				bytesLeft;
	const ODSShort		kBufferSize = 1000 ;
	ODSByte				buffer[kBufferSize] ;
	ODSLong				count;

	
	if (this->GetDraft()->ChangedFromPrev(ev))
	{
		// Externalize & Empty the working draft
		this->GetDraft()->Externalize(ev);
		this->GetDraft()->SaveToAPrevious (ev, kODNULL);
	}
	
	// Copy the Container buffer to the file

	tempFile = CreateTempFile(kODShellSignature,
						kODShellSignature, NULL);
	
	tempFile->SetPermission (fsRdWrPerm);
	
	CloneDraftToFileContainer (this->GetDraft(), tempFile);
	//this->CloneToFileContainer(tempFile);
	tempFile->Open();
	
	// Write the file container stream
	bytesLeft = tempFile->GetEndOfFile();
	length = bytesLeft;
	
	while( bytesLeft > 0 )
	{
		count = bytesLeft < kBufferSize ? bytesLeft: kBufferSize ;
		tempFile->Read( buffer, &count );
		file->Write( buffer, &count );
		bytesLeft -= count ;
	}

	tempFile->Close();
	tempFile->Delete();
	ODDeleteObject (tempFile);

	//fTmpCtrFile->Rename (file->GetFileSpec().name);

	return ((CASize)length);
	
}

//-------------------------------------------------------------------------

CASize CADocument::StreamToMemory(Handle*	containerHandle, CASize offset)
{
	Environment*		ev = gCASession->GetEV();
	ODSLong				length;
	ODPtr				bufferPtr;
	ODHandle			tempCtrHandle;
	
	if ((this->GetContainerType() == kCATempFileContainer) ||
	   (this->GetContainerType() == kCAInPlaceFileContainer))
	{
		DebugStr ("\pTrying to save a file container to memory");
	}
	
	else

	if (this->GetDraft()->ChangedFromPrev(ev))
	{
		// Externalize & Empty the working draft
		this->GetDraft()->Externalize(ev);
		this->GetDraft()->SaveToAPrevious (ev, kODNULL);
	}
	
	tempCtrHandle = ODNewHandle(0);
	CloneDraftToMemoryContainer (this->GetDraft(), &tempCtrHandle);
	
	ODLockHandle (tempCtrHandle);
	HLock (*containerHandle);
	bufferPtr = (Ptr) ((long)**containerHandle + offset);
	BlockMove (*tempCtrHandle, bufferPtr, ODGetHandleSize (tempCtrHandle));
	
	length = ODGetHandleSize (tempCtrHandle);

	ODUnlockHandle (tempCtrHandle);
	HUnlock (*containerHandle);
	
	return ((CASize)length);
	
}

void CADocument::AddLimboFrame (ODFrame* frame)
{
	Environment*		ev = gCASession->GetEV();

	if (!fLimboFrames)
		fLimboFrames = new OrderedCollection;
		
	fLimboFrames->AddLast (frame);
	
	TRACE (3, "Frame id %d added to CADocument, %x, limbo frame list", frame->GetID(ev), this);

}


//===========================================================================
#pragma mark '		CAClipboardDocument
//===========================================================================

//-------------------------------------------------------------------------

CAClipboardDocument::CAClipboardDocument (	ODClipboard* clipboard)
{
	
	fCloneKind		= 0;
	
	SetDocumentType(kCAClipboardDocument);
	SetContainerType(kCAInPlaceMemoryContainer);

}

//-------------------------------------------------------------------------

CAClipboardDocument::~CAClipboardDocument()
{

}


//-------------------------------------------------------------------------

void CAClipboardDocument::PrepareForClone(CACloneKind cloneKind)
{

	Environment*			ev = gCASession->GetEV();
	ODClipboard*			clipboard = gCASession->GetODSession()->GetClipboard(ev);

	fCloneKind			= cloneKind;
	fStreamOffset		= 0;

	if ((cloneKind == kCACloneCopy) || (cloneKind == kCACloneCut))
	{	
		ZeroScrap();
   		clipboard->Clear(ev);
   		fRootPartSU = clipboard->GetContentStorageUnit( ev );
   	}
	else if (cloneKind == kCAClonePaste)
	{	
	
   		fRootPartSU = clipboard->GetContentStorageUnit( ev );
   		if (fRootPartSU->Exists (ev, kODPropContents, kODNULL, 1))
			fRootPartSU->Focus (ev, kODPropContents, kODPosSame, kODNULL, 1, kODPosUndefined);
		
   	}

}


//-------------------------------------------------------------------------

Boolean  CAClipboardDocument::Close(void)
{
	Environment*		ev			= gCASession->GetEV();
	ODArbitrator*		arbitrator	= gCASession->GetArbitrator();;
	ODClipboard*		clipboard	= gCASession->GetODSession()->GetClipboard(ev);
	ODTypeToken			clipboardFocus;

	// Relinquish the clipboard focus
	clipboardFocus = gCASession->GetODSession()->Tokenize(ev, kODClipboardFocus);
	TempODFrame focusOwner = arbitrator->AcquireFocusOwner (ev, clipboardFocus);
	arbitrator->RelinquishFocus(ev, clipboardFocus, focusOwner);

	return (false);

}

//-------------------------------------------------------------------------
CADocumentType CAClipboardDocument::GetDocumentType(void)
{
	return kCAClipboardDocument;
}


//-------------------------------------------------------------------------

CASize CAClipboardDocument::StreamToFile(		PlatformFile* file)
{
	Environment*		ev = gCASession->GetEV();
	ODSLong				length;
	ODHandle			containerHandle;
	ODSLong				count;
	
	containerHandle = ODNewHandle(0);
	
	// Copy the clipboard memory container stream to the file

	CloneDraftToMemoryContainer (this->GetDraft(), &containerHandle);
	
	length = ODGetHandleSize (containerHandle);
	
	ODLockHandle (containerHandle);

	// Write the memory container stream
	count = length;
	file->Write( *containerHandle, &count );
	ODUnlockHandle (containerHandle);
	
	ODDisposeHandle (containerHandle);

	return ((CASize)length + sizeof (ODULong));
	
	
}


//===========================================================================
#pragma mark '		CADragDropDocument
//===========================================================================

//-------------------------------------------------------------------------

CADragDropDocument::CADragDropDocument(ODDragAndDrop* dragDrop)
{
	fRefCount 			= 0;
	fCloneKind			= 0;
	fDraggingFrames 	= new OrderedCollection;
	
	SetDocumentType(kCADragDropDocument);
	fInitiated = false;
	
}

//-------------------------------------------------------------------------

CADragDropDocument::~CADragDropDocument()
{
	ODDeleteObject (fDraggingFrames);
}


//-------------------------------------------------------------------------

void CADragDropDocument::Acquire(ItemReference theItem)
{

	Environment*			ev = gCASession->GetEV();
	ODDragAndDrop*			dragDrop = gCASession->GetODSession()->GetDragAndDrop(ev);
	ODDragLink*				theDragLink = kODNULL;
	ODMemDragItem*			theDragItem = kODNULL;
	LinkedListIterator* 	listIter	= kODNULL;

	++fRefCount;
	fDragItem = theItem;
	
   	if (theItem == 0)
   	{
   		dragDrop->Clear(ev);
   		fRootPartSU = dragDrop->GetContentStorageUnit( ev );
   		fInitiated = true;
   	}
	else
	{
		
		listIter = new LinkedListIterator(dragDrop->GetDragItemList(ev));
		
		for (theDragLink = (ODDragLink*)listIter->First(); theDragLink;
				theDragLink = (ODDragLink*)listIter->Next())
		{
			if ( (ItemReference)(theDragLink->fItem) == theItem)
			{
				fRootPartSU = theDragLink->fItem->fSU;
				break;
			}
		}
		
		ODDeleteObject (listIter);

	}

}

//-------------------------------------------------------------------------

void CADragDropDocument::PrepareForClone(CACloneKind cloneKind)
{

	Environment*			ev = gCASession->GetEV();
	ODDragAndDrop*			dragDrop = gCASession->GetODSession()->GetDragAndDrop(ev);
	ODDragLink*				theDragLink = kODNULL;
	ODMemDragItem*			theDragItem = kODNULL;
	LinkedListIterator* 	listIter	= kODNULL;

   	fCloneKind = cloneKind;
   	

}


//-------------------------------------------------------------------------

Boolean  CADragDropDocument::Close(void)
{

	Environment*				ev = gCASession->GetEV();
	ODDragAndDrop*				dragDrop = gCASession->GetODSession()->GetDragAndDrop(ev);
	OrderedCollectionIterator*	iter;
	ODFrame*					aFrame;
	
	--fRefCount;
	
	if (fRefCount == 0)
	{
		if (fInitiated)
		{
		
			dragDrop->Clear(ev);

			// Set dragging to false for dragged frames
	
			iter = fDraggingFrames->CreateIterator();
			for (aFrame = (ODFrame*)iter->First(); iter->IsNotComplete();
					aFrame = (ODFrame*)iter->Next())
			{
				aFrame->SetDragging (ev, kODFalse);
			}
			
			ODDeleteObject (iter);
	
			fDraggingFrames->RemoveAll ();

		}
		
		fInitiated = false;		// reset
		
	}
	
	return (false);

}

//-------------------------------------------------------------------------

void CADragDropDocument::AddDraggingFrame(ODFrame* frame)
{

	fDraggingFrames->AddLast (frame);

}

//-------------------------------------------------------------------------

void CADragDropDocument::SetDragging(void)
{

#ifdef DRAG_DEBUG
	if (fDragState != kDragLoading) DebugStr ("\p CADrag state error 4");
#endif
	
}

//-------------------------------------------------------------------------

CADocumentType CADragDropDocument::GetDocumentType(void)
{
	return kCADragDropDocument;
}




//===========================================================================
#pragma mark '		Utility Routines
//===========================================================================


//-------------------------------------------------------------------------

void CloneDraftToFileContainer	(ODDraft* srcDraft, PlatformFile* file)
{
	Environment*		ev			= gCASession->GetEV();
	ODSession*			session		= gCASession->GetODSession();;
	ODContainer*		container	= kODNULL;
	ODDocument*			document	= kODNULL;
	ODDraft*			draft		= kODNULL;
	ODStorageUnit*		props		= kODNULL;
	ODStorageUnitID		rootSUID;
	
	ODVolatile(draft);
	ODVolatile(document);
	ODVolatile(container);
	
	TempODStorageUnit tempSU = ODAcquireRootPartSUOfDraft (ev, srcDraft);
	rootSUID = tempSU->GetID(ev);

	TRY
	
		container = CreateFileContainer(ev, gCASession->GetODSession(),
						&(file->GetFileSpec()));
		document = container->AcquireDocument(ev, kODDefaultDocument);
		draft = document->AcquireBaseDraft(ev, kODDPExclusiveWrite);
		
		ODDraftKey draftKey = srcDraft->BeginClone(ev, draft, kODNULL, kODCloneAll);
		ODVolatile(draftKey);
		TRY
			rootSUID = srcDraft->Clone(ev, draftKey, rootSUID, kODNULLID, 0);	
		CATCH_ALL
			srcDraft->AbortClone(ev, draftKey);
		ENDTRY
		
		srcDraft->EndClone(ev, draftKey);
		
		props = draft->AcquireDraftProperties(ev);
		ODSetStrongSURefProp(ev, props, kODPropRootPartSU, kODStrongStorageUnitRef, rootSUID);
		props->Release(ev);
		
		draft->Externalize(ev);
		draft->Release(ev);
		document->Release(ev);
		container->Release(ev);

	CATCH_ALL
		ODReleaseObject(ev, draft);
		ODReleaseObject(ev, document);
		ODReleaseObject(ev, container);
	ENDTRY


}

//-------------------------------------------------------------------------
 
void CloneDraftToMemoryContainer	(ODDraft* srcDraft, ODHandle* containerHandle)
{
	Environment*		ev				= gCASession->GetEV();
	ODSession*			session			= gCASession->GetODSession();;
	ODContainer*		container		= kODNULL;
	ODDocument*			document		= kODNULL;
	ODDraft*			draft			= kODNULL;
	ODStorageUnit*		props			= kODNULL;
	ODStorageUnitID		rootSUID;

	ODVolatile(draft);
	ODVolatile(document);
	ODVolatile(container);
	ODVolatile(containerHandle);
	
	TempODStorageUnit tempSU = ODAcquireRootPartSUOfDraft (ev, srcDraft);
	rootSUID = tempSU->GetID(ev);
		
	// If the containerHandle is NULL, allocate a handle
	if (*containerHandle == kODNULL)
		*containerHandle = ODNewHandle(0);
		
	TRY
		container = CreateMemoryContainer(ev, session, *containerHandle, kODBentoMemoryContainer);
		document = container->AcquireDocument(ev, kODDefaultDocument);
		draft = document->AcquireBaseDraft(ev, kODDPExclusiveWrite);
		
		ODDraftKey draftKey = srcDraft->BeginClone(ev, draft, kODNULL, kODCloneAll);
		ODVolatile(draftKey);
		TRY
			rootSUID = srcDraft->Clone(ev, draftKey, rootSUID, kODNULLID, 0);	
		CATCH_ALL
			srcDraft->AbortClone(ev, draftKey);
		ENDTRY
		
		srcDraft->EndClone(ev, draftKey);

		props = draft->AcquireDraftProperties(ev);
		ODSetStrongSURefProp(ev, props, kODPropRootPartSU, kODStrongStorageUnitRef, rootSUID);
		props->Release(ev);

		draft->Externalize(ev);
		draft->Release(ev);
		document->Release(ev);
		container->Release(ev);

	CATCH_ALL
		ODReleaseObject(ev, draft);
		ODReleaseObject(ev, document);
		ODReleaseObject(ev, container);
		//ODDisposeHandle(containerHandle);
	ENDTRY

	//return containerHandle;

}

//-------------------------------------------------------------------------

PlatformFile* FileContainerFromMemoryContainer	(ODHandle memoryCtrStream)
{
	Environment*		ev				= gCASession->GetEV();
	ODSession*			session			= gCASession->GetODSession();;
	ODDraft*			baseDraft		= kODNULL;
	ODDraft*			topDraft		= kODNULL;
	PlatformFile*		containerFile;
	
	TempODContainer tempContainer = GetMemoryContainer(ev, gCASession->GetODSession(),
						memoryCtrStream, kODDefaultMemoryContainer);

	// Get the top draft
	TempODDocument tempDocument = tempContainer->AcquireDocument(ev,kODDefaultDocument);
	baseDraft = tempDocument->AcquireBaseDraft (ev, kODDPReadOnly);
	if (tempDocument->Exists (ev, 0, baseDraft, kODPosFirstAbove))
	{
		topDraft = tempDocument->AcquireDraft (ev, kODDPExclusiveWrite, 0, baseDraft,
										kODPosFirstAbove, kODTrue);
	}

	containerFile = CreateTempFile(kODShellSignature,
							kODShellSignature, NULL);
	containerFile->SetPermission (fsRdWrPerm);

	// Clone the draft into the container file
	CloneDraftToFileContainer (topDraft, containerFile);
	containerFile->Close();
	
	// Cleanup the temp memory container
	
	return (containerFile);

}

//-------------------------------------------------------------------------

ODHandle MemoryContainerFromFileContainer	(PlatformFile* fileCtr)
{
	Environment*		ev				= gCASession->GetEV();
	ODSession*			session			= gCASession->GetODSession();;
	ODDraft*			baseDraft		= kODNULL;
	ODDraft*			topDraft		= kODNULL;
	ODHandle			containerHandle;
	
	TempODContainer tempContainer = GetFileContainer(ev, gCASession->GetODSession(),
									&(fileCtr->GetFileSpec()));

	// Get the top draft
	TempODDocument tempDocument = tempContainer->AcquireDocument(ev,kODDefaultDocument);
	baseDraft = tempDocument->AcquireBaseDraft (ev, kODDPReadOnly);
	if (tempDocument->Exists (ev, 0, baseDraft, kODPosFirstAbove))
	{
		topDraft = tempDocument->AcquireDraft (ev, kODDPExclusiveWrite, 0, baseDraft,
										kODPosFirstAbove, kODTrue);
	}
	
	containerHandle = ODNewHandle (0);
	CloneDraftToMemoryContainer (topDraft, &containerHandle);
	
	// Cleanup the temp file container
	
	fileCtr->Close();
	
	return (containerHandle);


}
